@@ -20,13 +20,6 @@
 #'   else a symbolic loop will be used. Unrolling can speed-up a RNN, although 
 #'   it tends to be more memory-intensive. Unrolling is only suitable for short 
 #'   sequences.
-#' @param implementation one of {0, 1, or 2}. If set to 0, the RNN will use an 
-#'   implementation that uses fewer, larger matrix products, thus running faster
-#'   on CPU but consuming more memory. If set to 1, the RNN will use more matrix
-#'   products, but smaller ones, thus running slower (may actually be faster on 
-#'   GPU) while consuming less memory. If set to 2 (LSTM/GRU only), the RNN will
-#'   combine the input gate, the forget gate and the output gate into a single 
-#'   matrix, enabling more time-efficient parallelization on the GPU.
 #' @param kernel_initializer Initializer for the `kernel` weights matrix, used 
 #'   for the linear transformation of the inputs..
 #' @param recurrent_initializer Initializer for the `recurrent_kernel` weights 
@@ -57,7 +50,7 @@
 #'   
 #' @export
 layer_simple_rnn <- function(object, units, activation = "tanh", use_bias = TRUE, 
-                             return_sequences = FALSE, return_state = FALSE, go_backwards = FALSE, stateful = FALSE, unroll = FALSE, implementation = 0L,
+                             return_sequences = FALSE, return_state = FALSE, go_backwards = FALSE, stateful = FALSE, unroll = FALSE,
                              kernel_initializer = "glorot_uniform", recurrent_initializer = "orthogonal", bias_initializer = "zeros", 
                              kernel_regularizer = NULL, recurrent_regularizer = NULL, bias_regularizer = NULL, activity_regularizer = NULL, 
                              kernel_constraint = NULL, recurrent_constraint = NULL, bias_constraint = NULL, 
@@ -72,7 +65,6 @@ layer_simple_rnn <- function(object, units, activation = "tanh", use_bias = TRUE
     go_backwards = go_backwards,
     stateful = stateful,
     unroll = unroll,
-    implementation = as.integer(implementation),
     kernel_initializer = kernel_initializer,
     recurrent_initializer = recurrent_initializer,
     bias_initializer = bias_initializer,
@@ -122,7 +114,7 @@ layer_simple_rnn <- function(object, units, activation = "tanh", use_bias = TRUE
 #'     
 #' @export
 layer_gru <- function(object, units, activation = "tanh", recurrent_activation = "hard_sigmoid", use_bias = TRUE, 
-                      return_sequences = FALSE, return_state = FALSE, go_backwards = FALSE, stateful = FALSE, unroll = FALSE, implementation = 0L,
+                      return_sequences = FALSE, return_state = FALSE, go_backwards = FALSE, stateful = FALSE, unroll = FALSE,
                       kernel_initializer = "glorot_uniform", recurrent_initializer = "orthogonal", bias_initializer = "zeros", 
                       kernel_regularizer = NULL, recurrent_regularizer = NULL, bias_regularizer = NULL, activity_regularizer = NULL, 
                       kernel_constraint = NULL, recurrent_constraint = NULL, bias_constraint = NULL, 
@@ -138,7 +130,6 @@ layer_gru <- function(object, units, activation = "tanh", recurrent_activation =
     go_backwards = go_backwards,
     stateful = stateful,
     unroll = unroll,
-    implementation = as.integer(implementation),
     kernel_initializer = kernel_initializer,
     recurrent_initializer = recurrent_initializer,
     bias_initializer = bias_initializer,
@@ -167,6 +158,60 @@ layer_gru <- function(object, units, activation = "tanh", recurrent_activation =
 }
 
 
+#' Fast GRU implementation backed by [CuDNN](https://developer.nvidia.com/cudnn).
+#' 
+#' Can only be run on GPU, with the TensorFlow backend.
+#' 
+#' @inheritParams layer_simple_rnn
+#' 
+#' @family recurrent layers  
+#' 
+#' @section References: 
+#' - [On the Properties of Neural Machine Translation:
+#'   Encoder-Decoder Approaches](https://arxiv.org/abs/1409.1259) 
+#' - [Empirical
+#'   Evaluation of Gated Recurrent Neural Networks on Sequence
+#'   Modeling](http://arxiv.org/abs/1412.3555v1) 
+#' - [A Theoretically Grounded
+#'   Application of Dropout in Recurrent Neural
+#'   Networks](http://arxiv.org/abs/1512.05287)
+#'
+#' @export
+layer_cudnn_gru <- function(object, units,
+                            kernel_initializer = "glorot_uniform", recurrent_initializer = "orthogonal", bias_initializer = "zeros", 
+                            kernel_regularizer = NULL, recurrent_regularizer = NULL, bias_regularizer = NULL, activity_regularizer = NULL, 
+                            kernel_constraint = NULL, recurrent_constraint = NULL, bias_constraint = NULL,
+                            return_sequences = FALSE, return_state = FALSE, stateful = FALSE,
+                            input_shape = NULL, batch_input_shape = NULL, batch_size = NULL, 
+                            dtype = NULL, name = NULL, trainable = NULL, weights = NULL) {
+  args <- list(
+    units = as.integer(units),
+    kernel_initializer = kernel_initializer,
+    recurrent_initializer = recurrent_initializer,
+    bias_initializer = bias_initializer,
+    kernel_regularizer = kernel_regularizer,
+    recurrent_regularizer = recurrent_regularizer,
+    bias_regularizer = bias_regularizer,
+    activity_regularizer = activity_regularizer,
+    kernel_constraint = kernel_constraint,
+    recurrent_constraint = recurrent_constraint,
+    bias_constraint = bias_constraint,
+    return_sequences = return_sequences,
+    return_state = return_state,
+    stateful = stateful,
+    input_shape = normalize_shape(input_shape),
+    batch_input_shape = normalize_shape(batch_input_shape),
+    batch_size = as_nullable_integer(batch_size),
+    dtype = dtype,
+    name = name,
+    trainable = trainable,
+    weights = weights
+  )
+  
+  create_layer(keras$layers$CuDNNGRU, object, args)
+}
+
+
 #' Long-Short Term Memory unit - Hochreiter 1997.
 #' 
 #' For a step-by-step description of the algorithm, see [this
@@ -190,7 +235,7 @@ layer_gru <- function(object, units, activation = "tanh", recurrent_activation =
 #'     
 #' @export
 layer_lstm <- function(object, units, activation = "tanh", recurrent_activation = "hard_sigmoid", use_bias = TRUE, 
-                       return_sequences = FALSE, return_state = FALSE, go_backwards = FALSE, stateful = FALSE, unroll = FALSE, implementation = 0L,
+                       return_sequences = FALSE, return_state = FALSE, go_backwards = FALSE, stateful = FALSE, unroll = FALSE,
                        kernel_initializer = "glorot_uniform", recurrent_initializer = "orthogonal", bias_initializer = "zeros", 
                        unit_forget_bias = TRUE, kernel_regularizer = NULL, recurrent_regularizer = NULL, bias_regularizer = NULL, 
                        activity_regularizer = NULL, kernel_constraint = NULL, recurrent_constraint = NULL, bias_constraint = NULL, 
@@ -206,7 +251,6 @@ layer_lstm <- function(object, units, activation = "tanh", recurrent_activation
     go_backwards = go_backwards,
     stateful = stateful,
     unroll = unroll,
-    implementation = as.integer(implementation),
     kernel_initializer = kernel_initializer,
     recurrent_initializer = recurrent_initializer,
     bias_initializer = bias_initializer,
@@ -235,6 +279,56 @@ layer_lstm <- function(object, units, activation = "tanh", recurrent_activation
   create_layer(keras$layers$LSTM, object, args)
 }
 
+#' Fast LSTM implementation backed by [CuDNN](https://developer.nvidia.com/cudnn).
+#' 
+#' Can only be run on GPU, with the TensorFlow backend.
+#'
+#' @inheritParams layer_lstm
+#' 
+#' @section References: 
+#' - [Long short-term memory](http://www.bioinf.jku.at/publications/older/2604.pdf) (original 1997 paper) 
+#' - [Supervised sequence labeling with recurrent neural networks](http://www.cs.toronto.edu/~graves/preprint.pdf) 
+#' - [A Theoretically Grounded Application of Dropout in Recurrent Neural Networks](http://arxiv.org/abs/1512.05287)
+#'  
+#' @family recurrent layers  
+#' 
+#' @export
+layer_cudnn_lstm <- function(object, units, 
+                             kernel_initializer = "glorot_uniform",  recurrent_initializer = "orthogonal", 
+                             bias_initializer = "zeros",  unit_forget_bias = TRUE,
+                             kernel_regularizer = NULL, recurrent_regularizer = NULL, bias_regularizer = NULL, activity_regularizer = NULL, 
+                             kernel_constraint = NULL, recurrent_constraint = NULL, bias_constraint = NULL,
+                             return_sequences = FALSE, return_state = FALSE, stateful = FALSE,
+                             input_shape = NULL, batch_input_shape = NULL, batch_size = NULL, 
+                             dtype = NULL, name = NULL, trainable = NULL, weights = NULL) {
+  args <- list(
+    units = as.integer(units),
+    kernel_initializer = kernel_initializer,
+    recurrent_initializer = recurrent_initializer,
+    bias_initializer = bias_initializer,
+    unit_forget_bias = unit_forget_bias,
+    kernel_regularizer = kernel_regularizer,
+    recurrent_regularizer = recurrent_regularizer,
+    bias_regularizer = bias_regularizer,
+    activity_regularizer = activity_regularizer,
+    kernel_constraint = kernel_constraint,
+    recurrent_constraint = recurrent_constraint,
+    bias_constraint = bias_constraint,
+    return_sequences = return_sequences,
+    return_state = return_state,
+    stateful = stateful,
+    input_shape = normalize_shape(input_shape),
+    batch_input_shape = normalize_shape(batch_input_shape),
+    batch_size = as_nullable_integer(batch_size),
+    dtype = dtype,
+    name = name,
+    trainable = trainable,
+    weights = weights
+  )
+  
+  create_layer(keras$layers$CuDNNLSTM, object, args)
+}
+
 
 
 
