@@ -3,7 +3,7 @@ context("custom-layers")
 source("utils.R")
 
 # Custom layer class
-CustomLayer <- R6::R6Class("KerasLayer",
+CustomLayer <- R6::R6Class("CustomLayer",
                                   
   inherit = KerasLayer,
   
@@ -27,6 +27,7 @@ CustomLayer <- R6::R6Class("KerasLayer",
     },
     
     call = function(x, mask = NULL) {
+      self$add_loss(list(5))
       K <- backend()
       K$dot(x, self$kernel)
     },
